'''
BusyAnimation.py

An alternative to wx.BusyInfo or an indeterminate progress dialog, it makes use of ActiveBitmap.py
 to display a gif image or a list of images, as an animation, rather than use a gauge or simple text
 to indicate a delay in processing.

You may request that the Busy animation self terminates after a preset number of seconds, terminate it
 by calling Stop or the Busy window can simply closed by the user.
You may request that it watches an external process id and terminate when that process terminates.
 [this depends upon having the psutil library loaded]

You may choose to display an elapsed time, an elapsed time and time remaining, just time remaining or nothing,
 depending on whether the animation is self terminating or not.
Times are displayed in standard format, adjusted for the presence of an element, so hh:mm:ss.ff, mm:ss.ff or ss.ff
 i.e. 3662.55 seconds = 01:01:02.55 | 78.55 seconds = 01:18.55 | 10.55 seconds = 10.55

images for use in the animation are either a list of wx.Bitmaps e.g.
    images = [wx.Bitmap('1.png'),wx.Bitmap('2.png'),wx.Bitmap('3.png'),]
 or a gif image e.g:
    images = ['hourglass.gif',]

BusyAnimation can be loaded as a module, where you are responsible for handling Yielding to wx.App,
 as with any normal long running process or you may call it as a subprocess with the parameters you require
 and it will run standalone.
 e.g.
    subprocess.Popen([sys.executable, 'BusyAnimation.py', 'loading.gif,', '100', '1', '15', '-1'],).pid
    subprocess.Popen([sys.executable, 'BusyAnimation.py', "1.png, 2.png, 3.png", '100', '1', '15', '-1'],).pid

 Obviously, in this mode you cannot interact with it, other than kill it manually.

****
**** Note: I'm not sure if this subprocess route breaks any fundamental rules, so I'd appreciate any feedback if it does
****       Testing was on Linux only, so using psutil and signal modules may act differently on other platforms
****

Author:     J Healey
Created:    12/07/2023
Copyright:  J Healey - 2023
License:    GPL 3 or any later version
Version:    1.0.0
Email:      <rolfofsaxony@gmx.com>
Written & tested: Linux, Python 3.8.10, wxPython 4.2.1 gtk3 (phoenix) wxWidgets 3.2.2.1,
                  Pillow version 9.5.0, psutil version 5.5.1, ActiveBitmap.py version 1.2.2

ActiveBitmap.py version 1.2.2+ is available at:
    https://discuss.wxpython.org/t/activebitmaps-or-quasi-animation/36491

    Default class constructor.

        @param parent:  Parent window.
        @param id:      identifier. A value of -1 indicates a default value.
        @param title:   Title of the Busy Frame
        @param pos:     Position. If the position (-1, -1) is specified
                        then a default position is chosen.
        @param size:    Size of the Busy Frame.
        @param style:   wx.Border style default wx.STAY_ON_TOP
        @param msg:     Text to display as a message
        @param images:  A list containing zero or more wx.Bitmap entries that will by cycled through when activated
                        A list containg a single Gif filename is a valid entry
        @param speed:   The speed at which to cycle the images in the animation in milliseconds. (Default 50ms)
        @param elapse:  -1 (default) no elapsed time message
                         0 show elapsed time
                         1 show elapsed time and time remaining, if appropriate
                         2 show just time remaining, if appropriate
        @param timer:   self cancel animation in this number of seconds (default -1 do not self cancel)
        @param pid_lock: the number of an external process pid to watch (default -1 - no pid to watch)
                         when that process terminates, the busy animation will terminate too.
        @param bgclr:   Override Background colour of Busy widget. Format: html hex string '#000000' Default '' - none
        @param fgclr:   Override Foreground colour of Busy widget. Format: html hex string '#ffffff' Default '' - none
        @param xpos:    Override x screen position (default -1)
        @param ypos:    Override y screen position (default -1)
        @param width:   Override Busy widget width (default -1)
        @param height:  Override Busy widget height (default -1)

        By default, if the Busy widget has a parent it will use the foreground and background colours of the parent
         the parameters bgclr and fgclr allow for that to be overridden and in cases where the Busy widget
         is being run as an independent process (no parent), to allow it to be coloured.

        if xpos and ypos are both set to -2 the Busy widget will be Centred on the screen.
        if xpos or ypos are set to -2 the Busy widget will be Centred Horizontally or Vertically respectively.

    Functions:
        Stop        Stops the busy animation and closes the frame

        IsRunning() Return True or False if the control is Active

Example program:

import wx
from BusyAnimation import BusyAnimation

class Demo(wx.Frame):
    def __init__(self, parent):
        wx.Frame.__init__(self, parent, -1, "BusyAnimation Demonstration", size=(320, 450))
        self.Show()
        self.busy = BusyAnimation(self, wx.ID_ANY, title="Busy", size = (260, 260), msg="Processing...", \
                                images = ['loading.gif',], \
                                speed=100, elapse=1, timer = 10, pid_lock=-1)

if __name__ == '__main__':
    app = wx.App()
    frame = Demo(None)
    app.MainLoop()

'''
import wx
from ActiveBitmap import ActiveBitmap
import time
import sys
try:
    import psutil
    pid_check = True
except Exception:
    pid_check = False


class BusyAnimation(wx.Frame):
    def __init__(self, parent, id, title="Busy", pos = wx.DefaultPosition,
                 size = (300, -1), style=wx.STAY_ON_TOP, msg='', images=[], \
                 speed=50, elapse=-1, timer=-1, pid_lock=-1, \
                 bgclr="", fgclr="", xpos=-1, ypos=-1, width=-1, height=-1):
        wx.Frame.__init__(self, parent, wx.ID_ANY, title, pos=pos, size=size)
        self.pid_lock = pid_lock
        self.timeout = timer
        self.elapse = elapse
        if parent:
            self.SetBackgroundColour(parent.GetBackgroundColour())
            self.SetForegroundColour(parent.GetForegroundColour())
        if bgclr:
            self.SetBackgroundColour(bgclr)
        if fgclr:
            self.SetForegroundColour(fgclr)
        if xpos > 0 or ypos > 0:
            self.SetPosition((xpos, ypos))
        if xpos == -2 and ypos == -2:
            self.Centre()
        elif xpos == -2:
            self.Centre(wx.HORIZONTAL)
        elif ypos == -2:
            self.Centre(wx.VERTICAL)
        if width > 0 or height > 0:
            self.SetSize((width, height))
        self.ctrl = ActiveBitmap(self, -1, pos=(10, 10), bitmap_list=images)
        txt = wx.StaticText(self, -1, msg)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.ctrl, 0, wx.CENTRE|wx.ALL, 4)
        sizer.Add(txt, 0, wx.CENTRE|wx.ALL, 4)

        # Lock Busy to wait for a process to end - override elapse and timer parameters
        if pid_lock > 0 and pid_check:
            if psutil.pid_exists(self.pid_lock):
                p = psutil.Process(self.pid_lock)
                msg = 'Waiting for process: '+str(pid_lock)
                self.pid = wx.StaticText(self, -1, msg)
                self.pname = wx.StaticText(self, -1, "[ "+p.name()+" - "+p.status()+" ]")
                sizer.Add(self.pid, 0, wx.CENTRE|wx.ALL, 4)
                sizer.Add(self.pname, 0, wx.CENTRE|wx.ALL, 4)
                self.timeout = -1
                self.elapse = 0
            else:
                msg = "Error: Probably the requested pid "+str(pid_lock)+" does not exist"
                wx.MessageBox(msg, 'Error', wx.OK | wx.ICON_ERROR)
        elif pid_lock > 0 and not pid_check:
                msg = "Error: python package 'psutil' Not found on this machine"
                wx.MessageBox(msg, 'Error', wx.OK | wx.ICON_ERROR)
        self.Bind(wx.EVT_CLOSE, self.Stop)

        self.ctrl.SetTimerPeriod(speed)
        self.ctrl.Activate(True)

        self.elapsetimer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnElapse, self.elapsetimer)
        self.lifetimer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.Stop, self.lifetimer)
        # Automatically Cancel Busy after x many seconds
        if self.timeout > 0:
            timeout = self.timeout*1000
            self.lifetimer.StartOnce(timeout)
            self.ctrl.Activate(False, delay=timer) # Stop animation after timer seconds

        if self.elapse >= 0:
            self.time = time.time()
            self.elapsetimer.Start(100)

        # Show elapsed time
        if self.elapse >= 0:
            self.elapsed_time = wx.StaticText(self, -1, 'Elapsed: 0')
            sizer.Add(self.elapsed_time, 1, wx.ALL | wx.CENTRE, 4)

        self.SetSizer(sizer)
        self.Layout()
        self.Show()

    def OnElapse(self, event): # Elapsed time
        if self.pid_lock > 0:
            if not psutil.pid_exists(self.pid_lock): # simplest test for process
                self.Stop(None)
                return
        if self.pid_lock > 0:
            try:
                p = psutil.Process(self.pid_lock)
            except Exception:
                self.Stop(None)
                return
            stat = p.status()
            if stat == psutil.STATUS_ZOMBIE: # Test for zombie process
                self.Stop(None)
            msg = 'Waiting for process: '+str(self.pid_lock)
            self.pid.SetLabel(msg)
            msg = "[ "+" ".join(p.cmdline())+" - "+p.status()+" ]"
            self.pname.SetLabel(msg)

        t = time.time() - self.time
        emins, esecs = divmod(self.timeout - t, 60)
        eh, em = divmod(emins, 60)
        epmins, epsecs = divmod(t, 60)
        eph, epm = divmod(epmins, 60)
        if self.elapse > 1 and self.timeout > 0:
            if eh:
                etime = "Remaining: "+"{:02d}:{:02d}:{:05.2f}".format(int(eh),int(em),esecs)
            elif em:
                etime = "Remaining: "+"{:02d}:{:05.2f}".format(int(em),esecs)
            else:
                etime = "Remaining: "+"{:0.2f}".format(esecs)
            #etime = "Remaining: "+"{:0.2f}".format(self.timeout - t)
        elif self.elapse > 0 and self.timeout > 0:
            if eph:
                e1 = "Elapsed: "+"{:02d}:{:02d}:{:05.2f}".format(int(eph),int(epm),epsecs)
            elif epm:
                e1 = "Elapsed: "+"{:02d}:{:05.2f}".format(int(epm),epsecs)
            else:
                e1 = "Elapsed: "+"{:0.2f}".format(epsecs)
            if eh:
                e2 = " - Remaining: "+"{:02d}:{:02d}:{:05.2f}".format(int(eh),int(em),esecs)
            elif em:
                e2 = " - Remaining: "+"{:02d}:{:05.2f}".format(int(em),esecs)
            else:
                e2 = " - Remaining: "+"{:0.2f}".format(esecs)
            etime = e1+e2
            #etime = "Elapsed: "+"{:0.2f}".format(t)+" - Remaining: "+"{:0.2f}".format(self.timeout - t)
        else:
            if eph:
                etime = "Elapsed: "+"{:02d}:{:02d}:{:05.2f}".format(int(eph),int(epm),epsecs)
            elif epm:
                etime = "Elapsed: "+"{:02d}:{:05.2f}".format(int(epm),epsecs)
            else:
                etime = "Elapsed: "+"{:0.2f}".format(epsecs)
            #etime = "Elapsed: "+"{:0.2f}".format(t)
        self.elapsed_time.SetLabel(etime)
        self.Layout()

    def Stop(self, event=None):
        self.ctrl.Stop()
        self.elapsetimer.Stop()
        self.lifetimer.Stop()
        self.Destroy()

    def IsRunning(self):
        return self.ctrl.IsRunning()


class Demonstration(wx.Frame):
    def __init__(self):
        wx.Frame.__init__(self, None, title='MainFrame')
        self.SetBackgroundColour("grey80")
        txt = wx.StaticText(self, -1, "Testing 1 2 3 ....")
        self.sbutton = wx.Button(self, -1, label='Load Busy')
        self.ebutton = wx.Button(self, -1, label='Stop Busy')
        self.Bind(wx.EVT_BUTTON, self.OnStart, self.sbutton)
        self.Bind(wx.EVT_BUTTON, self.OnStop, self.ebutton)
        self.Bind(wx.EVT_CLOSE, self.OnQuit)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(txt, 0, wx.CENTRE|wx.ALL, 4)
        sizer.Add(self.sbutton, 0, wx.CENTRE|wx.ALL, 4)
        sizer.Add(self.ebutton, 0, wx.CENTRE|wx.ALL, 4)
        self.ebutton.Disable()
        self.SetSizer(sizer)
        self.Layout()
        self.Show()

    def OnStart(self, event):
        self.sbutton.Disable()
        self.ebutton.Enable()
        self.busy = BusyAnimation(self, wx.ID_ANY, title="Busy", size = (260, 260), msg="Processing...", \
                                #images = [wx.Bitmap('1.png'),wx.Bitmap('2.png'),wx.Bitmap('3.png'),], \
                                images = ['loading.gif',], \
                                speed=100, elapse=1, timer = 10, pid_lock=-1)
        while True:
            if not self.busy.IsRunning(): # Busy can be cancelled by clicking Frame Close
                break
            time.sleep(.5)
            wx.GetApp().Yield()

        try:
            self.ebutton.Disable()
            self.sbutton.Enable()
        except Exception:
            pass

    def OnStop(self, event):
        try:
            if self.busy.Stop():
                del self.busy
        except Exception:
            pass
        self.ebutton.Disable()
        self.sbutton.Enable()

    def OnQuit(self, event):
        self.OnStop(None)
        self.Destroy()


if __name__ == '__main__':
    app = wx.App()
    # If called from the cmdline or via subprocess with parameters - load directly, run standalone
    # e.g.  python3 BusyAnimation.py 'loading.gif,' 100 1 15 -1
    # or    python3 BusyAnimation.py '1.png,2.png,3.png' 100 1 15 -1
    # or    self.proc = subprocess.Popen([sys.executable, 'BusyAnimation.py', 'loading.gif,', '100', '1', '-1', '-1'],).pid
    #
    # to add a background colour for Busy add a 7th parameter specifying a html hex colour e.g. '#000000' for black
    # to add a foreground colour for Busy add an 8th parameter specifying a html hex colour e.g. '#ffffff' for white
    # if you require a foreground colour but no background colour pass the 7th parameter as an empty string i.e. ''
    # example:
    #       self.proc = Popen([sys.executable, 'BusyAnimation.py', 'progress.gif', '100', '2', '65', '-1', \
    #                                          '#000000', '#ff0000'],).pid
    #
    if len(sys.argv) > 1 and len(sys.argv) < 13:
        lst = sys.argv[1].split(',')
        image_lst = [x for x in lst if x]                               # clear out empty items
        if len(image_lst) > 1:                                          # static image list
            images = [wx.Bitmap(i.strip()) for i in image_lst]
        else:                                                           # a gif file
            images = [i.strip() for i in image_lst]
        if len(sys.argv) > 2:
            speed=int(sys.argv[2])
        else:
            speed=50
        if len(sys.argv) > 3:
            elapse=int(sys.argv[3])
        else:
            elapse=-1
        if len(sys.argv) > 4:
            timer = int(sys.argv[4])
        else:
            timer=-1
        if len(sys.argv) > 5:
            pid_lock = int(sys.argv[5])
        else:
            pid_lock=-1
        if len(sys.argv) > 6:
            bgclr = sys.argv[6]
        else:
            bgclr = ""
        if len(sys.argv) > 7:
            fgclr = sys.argv[7]
        else:
            fgclr = ""
        if len(sys.argv) > 8:
            xpos = int(sys.argv[8])
        else:
            xpos = -1
        if len(sys.argv) > 9:
            ypos = int(sys.argv[9])
        else:
            ypos = -1
        if len(sys.argv) > 10:
            width = int(sys.argv[10])
        else:
            width = -1
        if len(sys.argv) > 11:
            height = int(sys.argv[11])
        else:
            height = -1
        BusyAnimation(None, wx.ID_ANY, title="Busy", size = (260, 260), msg="Processing...", \
                                images = images, speed=speed, elapse=elapse, timer=timer, pid_lock=pid_lock, \
                                bgclr=bgclr, fgclr=fgclr, xpos=xpos, ypos=ypos, width=width, height=height)

    else:
        frame = Demonstration()

    app.MainLoop()
